from typing import Sequence

import h5py
import hdf5plugin  # noqa: F401
import numpy as np
from ewokscore import Task
from ewoksfluo.tasks import xrf_results
from silx.io import h5py_utils

# from ewokscore import Task


# from ewoksfluo.io.hdf5 import ReadHdf5File


# from ewoksfluo.io.hdf5 import ReadHdf5File
# from ewoksfluo.tasks import xrf_results


class WeightedSumResults(
    Task,
    input_names=[
        "xrf_norm_uris",
        "xrf_fit_uris",
        "bliss_scan_uri",
        "detector_names",
        "output_root_uri",
    ],
    optional_input_names=["detector_normalization_template"],
    output_names=["xrf_results_uri", "bliss_scan_uri", "output_root_uri"],
):
    """Add single-scan XRF results of multiple detectors"""

    def _read_xrf_results(self, uri):
        res = dict()

        fit_filename, fit_h5path = uri.split("::")

        fit_h5path += "/results"

        #        with ReadHdf5File(fit_filename) as h5file:

        with h5py_utils.open_item(fit_filename, "/") as h5file:
            try:
                xrf_results_group = h5file[fit_h5path]
            except KeyError:
                raise KeyError(
                    f"HDF5 path not found: '{fit_h5path}' in file '{fit_filename}'"
                )
            if not isinstance(xrf_results_group, h5py.Group):
                raise TypeError(
                    f"Expected HDF5 Group at '{fit_h5path}', but got {type(xrf_results_group)}"
                )

            if "massfractions" not in xrf_results_group:
                raise KeyError(f"'massfractions' group missing under '{fit_h5path}'")
            param_group = xrf_results_group["massfractions"]
            if not isinstance(param_group, h5py.Group):
                raise TypeError(
                    f"Expected HDF5 Group for 'massfractions', but got {type(param_group)}"
                )

            for dset_name, dset in param_group.items():
                if not xrf_results.is_peak_area(dset):
                    continue

                res[dset_name] = np.array(param_group[dset_name][()])

        return res

    def _dict_op(self, fun, a, b):
        ka = set(a.keys())
        kb = set(b.keys())

        s = dict()

        for k in set.intersection(ka, kb):
            s[k] = fun(a[k], b[k])

        return s

    def dict_op(self, fun, *op):
        if len(op) <= 1:
            return op

        res = self._dict_op(fun, op[0], op[1])

        for i in range(2, len(op)):
            res = self._dict_op(fun, res, op[i])

        return res

    def dict_plus(self, *op):
        return self.dict_op(lambda a, b: a + b, *op)

    def dict_mult(self, *op):
        return self.dict_op(lambda a, b: a * b, *op)

    def run(self) -> None:
        params = {**self.get_input_values()}

        xrf_norm_uris: Sequence[str] = params["xrf_norm_uris"]
        xrf_fit_uris: Sequence[str] = params["xrf_fit_uris"]
        bliss_scan_uri: str = params["bliss_scan_uri"]
        output_root_uri: str = params["output_root_uri"]

        if len(xrf_norm_uris) < 1:
            raise ValueError("Expected at least 1 detector to sum")

        if len(xrf_norm_uris) != len(xrf_fit_uris):
            raise ValueError(
                "Expected the same number of elements in _norm_ and _fit_ arrays."
            )

        summed_fit = None
        summed_prod = None

        config = {"xrf_norm_uris": xrf_norm_uris, "xrf_fit_uris": xrf_fit_uris}

        for norm_uri, fit_uri in zip(xrf_norm_uris, xrf_fit_uris):
            norm_data = self._read_xrf_results(norm_uri)
            fit_data = self._read_xrf_results(fit_uri)

            if summed_fit is None:
                summed_fit = fit_data
            else:
                summed_fit = self.dict_plus(summed_fit, fit_data)

            if summed_prod is None:
                summed_prod = self.dict_mult(norm_data, fit_data)
            else:
                summed_prod = self.dict_plus(
                    summed_prod, self.dict_mult(norm_data, fit_data)
                )

        weighted = self._dict_op(lambda a, b: a / b, summed_prod, summed_fit)

        xrf_results.save_xrf_results(
            output_root_uri,
            "weighted_ngmm2",
            config,
            None,
            None,
            weighted,
        )

        self.outputs.bliss_scan_uri = bliss_scan_uri
        self.outputs.output_root_uri = output_root_uri
        self.outputs.xrf_results_uri = output_root_uri + "/results"



from configparser import ConfigParser

import numpy as np
from ewokscore import Task
from ewoksfluo.tasks import nexus_utils
from ewoksfluo.tasks.hdf5_utils import create_hdf5_link
from ewoksfluo.tasks.hdf5_utils import link_bliss_scan
from silx.io import h5py_utils


class NormalizeCurrent(
    Task,
    input_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "current_counter",
        "reference_uri",
        "dwelltime_counter",
        "ref_current_counter",
        "diode_factor",
    ],
    optional_input_names=["gamma_coef", "min_ref", "max_ref", "exp_default"],
    output_names=["bliss_scan_uri", "output_root_uri"],
):

    def run(self):
        start_time = nexus_utils.now()
        bliss_scan_uri = self.inputs.bliss_scan_uri
        output_root_uri = self.inputs.output_root_uri
        scanuris = bliss_scan_uri.split("::")

        filename = scanuris[0]
        scanno = scanuris[1]

        counter_template = f"/{scanno}/instrument/%s/data"

        ref_scanuris = self.inputs.reference_uri.split("::")

        ref_filename = ref_scanuris[0]
        ref_scanno = ref_scanuris[1]

        ref_counter_template = f"/{ref_scanno}/instrument/%s/data"

        # Wait for scan to be finished first, up to one hour.
        with h5py_utils.open_item(
            filename,
            f"/{scanno}/end_time",
            retry_invalid=True,
            retry_timeout=3600,
            retry_period=5,
        ) as end_time:
            print(f"Scan finished at {end_time[()]}")

        with h5py_utils.open_item(
            ref_filename, ref_counter_template % self.inputs["ref_current_counter"]
        ) as ds:  # type: ignore[reportGeneralTypeIssues]
            # ref_epoch  = np.array(fd[ref_counter_template%self.inputs.get("ref_epoch_counter", "epoch")])
            ref_values = np.array(ds)

        with h5py_utils.open_item(filename, "/") as fd:  # type: ignore[reportGeneralTypeIssues]
            # epoch  = np.array(fd[ref_counter_template%self.inputs.get("data_epoch_counter", "epoch_trig")])
            values = np.array(fd[counter_template % self.inputs["current_counter"]])
            dwelltime = np.mean(fd[counter_template % self.inputs["dwelltime_counter"]])

        min_ref = self.get_input_value("min_ref", 1e8)
        max_ref = self.get_input_value("max_ref", 1e15)

        diode_factor = self.inputs.diode_factor * 1e12  # Convert ph/pA to ph/A

        ref_msk = np.logical_and(
            ref_values * diode_factor > min_ref, ref_values * diode_factor < max_ref
        )  # When shutter was open and not saturated

        #        ref_epoch = ref_epoch[ref_msk]
        ref_values = ref_values[ref_msk]

        gamma = np.mean(values) / (np.mean(ref_values) * dwelltime)
        gamma_coef = self.get_input_value("gamma_coef", 2.0)

        real_gamma_exp = np.log10(gamma / gamma_coef)
        gamma_exp = np.round(real_gamma_exp)

        gamma = gamma_coef * 10**gamma_exp

        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            start_time,
            process_config={"real_gamma_exp": real_gamma_exp, "gamma_exp": gamma_exp},
            default_levels=(scanno, "scaled_it"),
        ) as (process_group, already_existed):
            outentry = process_group.parent
            if not already_existed:
                link_bliss_scan(outentry, bliss_scan_uri, retry_timeout=0)

                nxdata = nexus_utils.create_nxdata(
                    process_group, "scaled_it", signal="data"
                )
                nxdata.attrs["interpretation"] = "spectrum"
                dset = nxdata.create_dataset("data", data=values / gamma * diode_factor)

                nxdetector = outentry["instrument"].create_group("scaled_it")
                nxdetector.attrs["NX_class"] = "NXdetector"
                create_hdf5_link(nxdetector, "data", dset)
                create_hdf5_link(outentry["measurement"], "scaled_it", dset)

            output_root_uri = f"{outentry.file.filename}::{outentry.name}"

        self.outputs.bliss_scan_uri = bliss_scan_uri
        self.outputs.output_root_uri = output_root_uri


class NormalizationFactorFromConfig(
    Task,
    input_names=["bliss_scan_uri", "xrf_results_uri", "config_file"],
    output_names=[
        "bliss_scan_uri",
        "xrf_results_uri",
        "counter_normalization_template",
    ],
):

    def get_from_cfg(self, cfg):
        cp = ConfigParser()
        cp.read(cfg)

        flux = cp.getfloat("concentrations", "flux")
        dwelltime = cp.getfloat("concentrations", "time")

        matrix = cp.get("attenuators", "Matrix")
        matdat = matrix.replace(" ", "").split(",")
        matrix_composition = matdat[1]
        matrix_density = float(matdat[2])
        matrix_thickness = float(matdat[3])
        print("INFO FROM CONFIG FILE")
        print("Config input flux: {0:g} (ph/s)".format(flux))
        print("Config dwell time: {0} (s)".format(dwelltime))
        print("Matrix composition: {0}".format(matrix_composition))
        print("Matrix density: {0} (g/cm**3)".format(matrix_density))
        print("Matrix thickness: {0} (cm)".format(matrix_thickness))
        areal_dens_ratio = matrix_density * matrix_thickness * 1e7  # ng/mm**2
        return flux, areal_dens_ratio, dwelltime

    def run(self):
        self.outputs.bliss_scan_uri = self.inputs.bliss_scan_uri
        self.outputs.xrf_results_uri = self.inputs.xrf_results_uri

        cfg_flux, areal_dens_ratio, cfg_dwelltime = self.get_from_cfg(
            self.inputs.config_file
        )

        factor = cfg_flux * cfg_dwelltime * areal_dens_ratio

        self.outputs.counter_normalization_template = (
            f"{factor:.04e}/<instrument/{{}}/data>"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov  1 17:01:47 2025

@author: blissadm
"""

import glob
import os

import h5py
from ewokscore import Task


class FluotomoMaster(
    Task,
    input_names=[
        "processed_dataset_folder",
    ],
    optional_input_names=["output_folder", "output_filename"],
    output_names=["output_master_file"],
):
    """Create a fluotomo master folder from all fluofit folders"""

    def run(self):
        inpath = self.inputs["processed_dataset_folder"]

        while inpath[-1] == "/":
            inpath = inpath[:-1]

        opath = self.inputs.get("output_folder")
        ofn = self.inputs.get("output_filename")

        if opath is None:
            opath = os.path.join(inpath, "fluotomo")

        datasetname = inpath.split("/")[-1]

        if ofn is None:
            ofn = datasetname + ".h5"

        os.makedirs(opath, exist_ok=True)
        masterfile = os.path.join(opath, ofn)
        self.outputs["output_master_file"] = masterfile

        with h5py.File(masterfile, "w") as ofd:
            for f in glob.glob(os.path.join(inpath, f"fluofit_*/{datasetname}.h5")):
                ss = f.split("/")[-2]
                if "scan" not in ss:
                    continue

                scanno = int(ss[-4:])
                # print(scanno)

                ofd[f"{scanno}.1"] = h5py.ExternalLink(f, f"/{scanno}.1")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul  6 20:10:52 2025

@author: blissadm
"""

import os

from ewokscore import Task
from pyicat_plus.client.main import IcatClient


class PublishProcessed(
    Task,
    input_names=["bliss_scan_file", "processed_folder", "icat_url"],
    optional_input_names=["beamline", "metadata", "dataset", "sample", "proposal"],
):
    """Publish processed data to data portal"""

    def run(self):
        raw_data = self._fix_data_path(self.inputs["bliss_scan_file"])
        data_folder = self._fix_data_path(self.inputs["processed_folder"])

        # Deduce from raw data path or get from optional inputs
        spath = raw_data.split("/")
        raw_data = os.path.dirname(raw_data)
        dataset = self.get_input_value(
            "dataset", spath[-2] if len(spath) >= 2 else None
        )
        sample = self.get_input_value(
            "sample", spath[-3] if len(spath) >= 3 else "sample"
        )
        beamline = self.get_input_value(
            "beamline", spath[-6] if len(spath) >= 6 else "id16a"
        )
        proposal = self.get_input_value(
            "proposal", spath[-7] if len(spath) >= 7 else None
        )

        metadata = self.get_input_value("metadata", {})

        if "Sample_name" not in metadata:
            metadata["Sample_name"] = sample

        print(
            self.inputs["bliss_scan_file"],
            raw_data,
            data_folder,
            dataset,
            sample,
            beamline,
            proposal,
        )

        client = IcatClient(metadata_urls=self.inputs["icat_url"])
        client.store_processed_data(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=data_folder,
            metadata=metadata,
            raw=raw_data,
        )

    def _fix_data_path(self, path):
        if path.startswith("/mnt/multipath-shares/data"):
            path = path.replace("/mnt/multipath-shares/data", "/data")

        return path



import os

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from ewokscore import Task
from PIL import Image
from silx.io import h5py_utils


class TiffExporterFromRegrid(
    Task,
    input_names=[
        "regrid_uri",
        "output_path",
    ],
    optional_input_names=[
        "output_prefix",
        "output_suffix",
        "output_gallery_path",
    ],
    output_names=[],
):
    def run(self):
        regrid_uri = self.inputs.regrid_uri
        output_path = self.inputs.output_path
        output_prefix = self.get_input_value("output_prefix", "IMG_")
        output_suffix = self.get_input_value("output_suffix", "")
        output_gallery_path = self.get_input_value("output_gallery_path", None)

        os.makedirs(output_path, exist_ok=True)
        if output_gallery_path is not None:
            os.makedirs(output_gallery_path, exist_ok=True)

        uris = regrid_uri.split("::")
        input_filename = uris[0]
        grp = uris[1]

        with h5py_utils.open_item(input_filename, "/") as fd:  # type: ignore[reportGeneralTypeIssues]
            _res = fd[grp]
            is_nxdata = "NX_class" in _res.attrs and _res.attrs["NX_class"] == "NXdata"

            while not is_nxdata:
                _res = _res[_res.attrs["default"]]
                is_nxdata = (
                    "NX_class" in _res.attrs and _res.attrs["NX_class"] == "NXdata"
                )

            grps = []
            if "signal" in _res.attrs:
                grps += [
                    _res.attrs["signal"],
                ]

            if "auxiliary_signals" in _res.attrs:
                grps += list(_res.attrs["auxiliary_signals"])

            kwargs = {}
            axes = []
            axnames = []
            if "axes" in _res.attrs:
                resolutions = []
                for a in _res.attrs["axes"]:
                    axnames += [
                        a,
                    ]
                    d = np.array(_res[a])
                    axes += [
                        d,
                    ]
                    resolutions += [
                        1e4 / float(np.mean(np.diff(d))),
                    ]

                kwargs["resolution_unit"] = 3
                kwargs["resolution"] = resolutions[::-1]
                axes = axes[::-1]
                axnames = axnames[::-1]

            cmap = "jet"

            for g in grps:
                data = np.array(_res[g], dtype=np.float32)
                im = Image.fromarray(data, mode="F")
                gg = g.replace(" ", "_")
                im.save(
                    os.path.join(
                        output_path, f"{output_prefix}{gg}{output_suffix}.tiff"
                    ),
                    **kwargs,
                )

                # ArraySave.save2DArrayListAsMonochromaticTiff([data], os.path.join(output_path, f"{output_prefix}{gg}.tiff"), dtype=np.float32)
                if output_gallery_path is not None:
                    norm = mpl.colors.Normalize(0, np.max(data))
                    f, ax = plt.subplots()
                    ax.pcolor(*axes, data, norm=norm, cmap=cmap, shading="nearest")
                    ax.invert_yaxis()
                    ax.set_aspect("equal", "box")

                    if len(axnames) == 2:
                        ax.set_xlabel(f"{axnames[0]} (um)")
                        ax.set_ylabel(f"{axnames[1]} (um)")

                    f.suptitle(g)
                    f.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
                        ax=ax,
                        label="Areal mass density ($ng/mm^2$)",
                    )
                    f.savefig(
                        os.path.join(
                            output_gallery_path,
                            f"{output_prefix}{gg}{output_suffix}.jpg",
                        )
                    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov  1 17:02:10 2025

@author: blissadm
"""

import os

import h5py
import imageio
import numpy as np
from ewokscore import Task


class AnimatedWebP(
    Task,
    input_names=["master_file", "detector_name"],
    output_names=["output_master_file"],
):
    """Create a fluotomo master folder from all fluofit folders"""

    def run(self):
        masterfile = self.inputs["master_file"]
        detname = self.inputs["detector_name"]
        self.outputs["output_master_file"]

        maps = {}
        shapes = {}
        somega = {}

        with h5py.File(masterfile, "r", locking=False) as fd:
            for e in fd:
                somega[e] = fd[e]["instrument/positioners/somega"][()]

                if "grid" not in fd[e]:
                    continue

                gridname = f"grid_{detname}_ng_mm2"

                if gridname not in fd[e]["grid"]:
                    continue

                group = fd[e]["grid"][gridname]["results/massfractions"]

                for g in group:
                    if len(group[g].shape) != 2:
                        continue

                    if g not in maps:
                        maps[g] = {}

                    if group[g].shape not in shapes:
                        shapes[group[g].shape] = 0

                    maps[g][e] = np.array(group[g])
                    shapes[group[g].shape] += 1

        mx = 0
        shp = None

        for sh in shapes:
            if shapes[sh] > mx:
                shp = sh
                mx = shapes[sh]

        processed_data_folder = os.path.dirname(masterfile)
        galleryfolder = os.path.join(processed_data_folder, "gallery")
        os.makedirs(galleryfolder, exist_ok=True)

        ssomega = {k: v for k, v in sorted(somega.items(), key=lambda item: item[1])}

        for elm, G in maps.items():
            filtered = []
            entries = []

            mn = None
            Mx = None

            for entry in ssomega:
                M = G[entry]

                if M.shape != shp:
                    continue

                filtered += [
                    M,
                ]
                entries += [
                    entry,
                ]

                mm = np.min(M)
                MM = np.max(M)

                if mn is None or mm < mn:
                    mn = mm

                if Mx is None or MM > Mx:
                    Mx = MM

            filtered = np.array(filtered)
            converted = np.array((filtered - mn) / (Mx - mn) * 255, dtype=np.uint8)

            imageio.v3.imwrite(
                os.path.join(galleryfolder, f"{elm}.webp"), converted, fps=5
            )



import os
import time

import matplotlib.pyplot as plt
import numpy as np
from ewokscore import Task
from PyMca5.PyMcaIO import ConfigDict
from PyMca5.PyMcaPhysics.xrf.ClassMcaTheory import ClassMcaTheory
from silx.io import h5py_utils


class AdvancedFitSumSingleDetector(
    Task,
    input_names=[
        "bliss_scan_uri",
        "detector_name",
        "config",
        "output_root_uri",
    ],
    optional_input_names=[
        "figure_filename",
        "batchconfig_suffix",
        "instrument_data_template",
        "batch_force",
        "waitForConfigFile",
        "retryPeriod",
        "retryN",
    ],
    output_names=[
        "bliss_scan_uri",
        "detector_name",
        "output_root_uri",
        "batch_config_filename",
    ],
):
    def run(self):
        input_uri = self.inputs.bliss_scan_uri
        output_uri = self.inputs.output_root_uri
        detector_name = self.inputs.detector_name
        config_file = self.inputs.config

        dettmpl = self.get_input_value("instrument_data_template", "instrument/{}/data")
        batch_suffix = self.get_input_value("batchconfig_suffix", None)
        batch_force = self.get_input_value(
            "batch_force",
            {
                "fit.stripflag": 0,
                "fit.escapeflag": 0,
                "fit.fitweight": 0,
                "fit.linearfitflag": 1,
            },
        )
        waitForConfigFile = self.get_input_value("waitForConfigFile", True)
        retryPeriod = self.get_input_value("retryPeriod", 3)
        retryN = self.get_input_value("retryN", 1200)

        figure_filename = self.get_input_value("figure_filename", None)

        uris = input_uri.split("::")
        filename = uris[0]
        datpath = uris[1] + "/" + dettmpl.format(detector_name)

        if waitForConfigFile:
            for i in range(retryN):
                if os.path.isfile(config_file):
                    break
                time.sleep(retryPeriod)
            else:
                raise RuntimeError(f"Config file {config_file} not found!")

        with h5py_utils.open_item(filename, datpath) as ds:  # type: ignore[reportGeneralTypeIssues]
            data = np.array(ds)

        batch_savefile = None

        if batch_suffix is not None:
            batch_savefile = config_file[:-4] + batch_suffix

            if not os.path.isfile(batch_suffix):
                cfg = ConfigDict.ConfigDict(filelist=config_file)

                for k in batch_force:
                    _c = cfg
                    kk = k.split(".")
                    for _k in kk[:-1]:
                        _c = _c[_k]

                    _c[kk[-1]] = batch_force[k]

                cfg.write(batch_savefile)

        data = np.sum(data, keepdims=True, axis=0) / data.shape[0]

        mcafit = ClassMcaTheory(config_file)
        mcafit.setData(x=np.arange(data.shape[1]), y=data)
        mcafit.estimate()
        p, fit = mcafit.startfit(digest=1)

        egy = fit["energy"]
        bkg = fit["continuum"]
        grps = fit["groups"]
        yfit = fit["yfit"]
        ydata = fit["ydata"]

        ouris = output_uri.split("::")
        out_filename = ouris[0]
        out_path = ouris[1]

        with h5py_utils.open_item(out_filename, "/", mode="a") as fd:  # type: ignore[reportGeneralTypeIssues]
            _res = fd
            for g in out_path.split("/"):
                _g = g.strip()
                if len(_g) == 0:
                    continue

                _res = _res.require_group(_g)
                _res.attrs["NX_class"] = "NXcollection"

            _res.attrs["NX_class"] = "NXdata"
            _res.attrs["signal"] = "fit"
            _res.attrs["auxiliary_signals"] = [
                "background",
                "spectrum",
            ] + grps
            _res.attrs["axes"] = ["energy"]

            egyds = _res.create_dataset("energy", data=egy)
            egyds.attrs["units"] = "keV"

            _res.create_dataset("fit", data=yfit)
            _res.create_dataset("spectrum", data=ydata)
            _res.create_dataset("background", data=bkg)

            for g in grps:
                _res.create_dataset(g, data=fit[f"y{g}"])

        if figure_filename is not None:
            figpath = os.path.dirname(figure_filename)
            os.makedirs(figpath, exist_ok=True)

            fig, ax = plt.subplots(figsize=(8, 5), layout="constrained")
            colors = plt.get_cmap("nipy_spectral")(np.linspace(0.1, 0.9, len(grps)))

            ax.plot(egy, ydata, "-", color="0.7", label="Data", linewidth=4)
            # ax.plot(egy, yfit, 'r-', label="Fit", linewidth=2)

            for c, g in zip(colors, grps):
                ax.plot(egy, fit[f"y{g}"] + bkg, "--", label=g, linewidth=1.5, color=c)

            ax.plot(egy, bkg, "b", label="Background", linewidth=2)

            fig.legend(loc="outside right upper")

            ax.set_yscale("log")
            ax.set_xlabel("Energy (keV)")

            fig.savefig(figure_filename)

        self.outputs.output_root_uri = output_uri
        self.outputs.detector_name = detector_name
        self.outputs.bliss_scan_uri = input_uri
        self.outputs.batch_config_filename = batch_savefile



from typing import Sequence

import h5py  # noqa: F401
import hdf5plugin  # noqa: F401
import numpy
from ewokscore import Task
from ewoksfluo.io.hdf5 import split_h5uri
from ewoksfluo.tasks import nexus_utils
from ewoksfluo.tasks.hdf5_utils import create_hdf5_link
from ewoksfluo.tasks.hdf5_utils import link_bliss_scan

# from ewoksfluo.io.hdf5 import ReadHdf5File
from ewoksfluo.tasks.math import eval_hdf5_expression
from ewoksfluo.tasks.math import format_expression_template
from silx.io import h5py_utils

DEFAULTS = {
    "xrf_spectra_uri_template": "instrument/{}/data",
    "detector_normalization_template": "1./<instrument/{}/live_time>",
    "output_detector_name": "mcasum",
}


class SumXrfSpectra(
    Task,
    input_names=[
        "bliss_scan_uri",
        "detector_names",
        "output_root_uri",
    ],
    optional_input_names=[
        "xrf_spectra_uri_template",
        "detector_normalization_template",
        "output_detector_name",
    ],
    output_names=[
        "bliss_scan_uri",
        "detector_name",
        "xrf_spectra_uri_template",
        "output_root_uri",
    ],
):
    """Add single-scan XRF spectra from multiple detectors"""

    def run(self):
        start_time = nexus_utils.now()
        params = {**DEFAULTS, **self.get_input_values()}

        bliss_scan_uri: str = params["bliss_scan_uri"]
        detector_names: Sequence[str] = params["detector_names"]
        xrf_spectra_uri_template: str = params["xrf_spectra_uri_template"]
        detector_normalization_template: str = params["detector_normalization_template"]
        output_root_uri: str = params["output_root_uri"]

        if len(detector_names) < 1:
            raise ValueError("Expected at least 1 detector to sum")

        _, scan_h5path = split_h5uri(bliss_scan_uri)

        sumdetector_name = params["output_detector_name"]

        print("BeforeWith")
        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            start_time,
            process_config={
                "detector_normalization_template": detector_normalization_template
            },
            default_levels=(scan_h5path, "sumspectra"),
        ) as (process_group, already_existed):
            outentry = process_group.parent
            print(outentry)
            if not already_existed:
                print("NOT EXIST, DOING")
                sum_spectra = _sum_spectra(
                    bliss_scan_uri,
                    xrf_spectra_uri_template,
                    detector_names,
                    detector_normalization_template,
                )
                print("AFTERSUM")

                link_bliss_scan(outentry, bliss_scan_uri, retry_timeout=0)

                print("AFTERLINK")

                nxdata = nexus_utils.create_nxdata(
                    process_group, "mcasum", signal="data"
                )
                nxdata.attrs["interpretation"] = "spectrum"

                #                Optimum compression but less portable....
                #                dset = nxdata.create_dataset("data", data=sum_spectra, chunks=(1, sum_spectra.shape[1]), compression=hdf5plugin.Bitshuffle(cname="lz4"))
                dset = nxdata.create_dataset(
                    "data",
                    data=sum_spectra,
                    chunks=(1, sum_spectra.shape[1]),
                    compression="gzip",
                    shuffle=True,
                )

                nxdetector = outentry["instrument"].create_group(sumdetector_name)
                nxdetector.attrs["NX_class"] = "NXdetector"
                create_hdf5_link(nxdetector, "data", dset)
                create_hdf5_link(outentry["measurement"], sumdetector_name, dset)

            output_root_uri = f"{outentry.file.filename}::{outentry.name}"

        self.outputs.bliss_scan_uri = output_root_uri
        self.outputs.detector_name = sumdetector_name
        self.outputs.xrf_spectra_uri_template = DEFAULTS["xrf_spectra_uri_template"]
        self.outputs.output_root_uri = output_root_uri


def _sum_spectra(
    bliss_scan_uri: str,
    xrf_spectra_uri_template: str,
    detector_names: str,
    detector_normalization_template: str,
) -> numpy.ndarray:
    """Add spectra from multiple detectors after normalizing the spectra to a common live time."""
    input_file, scan_h5path = split_h5uri(bliss_scan_uri)
    #    with ReadHdf5File(input_file) as h5file:
    with h5py_utils.open_item(input_file, "/") as h5file:
        sum_spectra = None
        scan_group = h5file[scan_h5path]

        for detector_name in detector_names:
            print(detector_name)
            xrf_spectra_dataset = scan_group[
                xrf_spectra_uri_template.format(detector_name)
            ]
            print(xrf_spectra_dataset)
            # assert isinstance(xrf_spectra_dataset, h5py.Dataset)
            xrf_spectra_data = xrf_spectra_dataset[()]

            if (
                detector_normalization_template is not None
            ):  # Allow to get an unnormalized sum
                weight_expression = format_expression_template(
                    detector_normalization_template, detector_name
                )
                weight = eval_hdf5_expression(bliss_scan_uri, weight_expression)
                weight = weight.reshape((len(weight), 1))

                xrf_spectra_data *= weight

            if sum_spectra is None:
                sum_spectra = xrf_spectra_data
            else:
                sum_spectra += xrf_spectra_data

    return sum_spectra



import h5py
import hdf5plugin  # noqa: F401
import numpy as np
from ewokscore import Task
from silx.io import h5py_utils


class SpectralRegrid(
    Task,
    input_names=["bliss_scan_uri", "output_root_uri", "counter_name"],
    optional_input_names=[],
    output_names=["bliss_scan_uri", "output_root_uri"],
):
    """Regrid raw spectrums"""

    def run(self):
        _uris = self.inputs.bliss_scan_uri.split("::")
        filename_src = _uris[0]
        scan_src = _uris[1]
        cntnam = self.inputs.counter_name

        fscanuri = f"{scan_src}/instrument/fscan_parameters"

        print(f"Waiting to open {filename_src}")
        with h5py_utils.open_item(filename_src, fscanuri) as fd:  # type: ignore[reportGeneralTypeIssues]
            print(f"{filename_src} open!")
            fastmot = fd["fast_motor"][()].decode()
            mode = fd["fast_motor_mode"][()].decode()
            fast_n = int(fd["fast_npoints"][()].decode())
            slow_n = int(fd["slow_npoints"][()].decode())

        with h5py_utils.open_item(
            filename_src, f"{scan_src}/instrument/{cntnam}"
        ) as fd:  # type: ignore[reportGeneralTypeIssues]
            cnt = fd["data"]
            cnt_shape = cnt.shape
            cnt_dtype = cnt.dtype

        A = np.arange(fast_n * slow_n)
        #        A.shape = (fast_n, slow_n)
        A.shape = (slow_n, fast_n)

        if fastmot.startswith("spz"):
            A = A.swapaxes(0, 1)

        if mode == "ZIGZAG":
            A[1::2, :] = A[1::2, :][:, ::-1]

        _uris = self.inputs.output_root_uri.split("::")
        filename = _uris[0]
        scan = _uris[1]

        print(f"Waiting to open {filename}")
        with h5py_utils.open_item(filename, "/", mode="a") as fd:  # type: ignore[reportGeneralTypeIssues]
            print(f"{filename} open!")
            grp = fd
            S = scan.split("/")

            for i, s in enumerate(S):
                if s == "":
                    continue

                grp = grp.require_group(s)

                if i == len(S) - 1:
                    grp.attrs.update(
                        {
                            "NX_class": "NXdata",
                            "signal": "data",
                            "interpretation": "image",
                        }
                    )
                else:
                    grp.attrs.update({"NX_class": "NXcollection", "default": S[i + 1]})

            virtual = False

            if not virtual:
                ds = np.empty(shape=(cnt_shape[1], *A.shape), dtype=cnt_dtype)

                #                with h5py_utils.open_item(filename_src, f"{scan_src}/instrument/{cntnam}/data", mode='r') as src:
                src = fd[f"{scan_src}/instrument/{cntnam}/data"]
                for i in range(cnt_shape[0]):
                    ds[(slice(None), *np.unravel_index(i, A.shape))] = src[A.flat[i]]

                if "data" in grp:
                    del grp["data"]

                grp.create_dataset(
                    "data",
                    data=ds,
                    chunks=(1, *A.shape),
                    shuffle=True,
                    compression="gzip",
                )

            else:
                layout = h5py.VirtualLayout(
                    shape=(
                        cnt_shape[1],
                        *A.shape,
                    ),
                    dtype=cnt_dtype,
                )
                layoutpymca = h5py.VirtualLayout(
                    shape=(
                        *A.shape,
                        cnt_shape[1],
                    ),
                    dtype=cnt_dtype,
                )
                vsource = h5py.VirtualSource(
                    filename_src,
                    f"{scan_src}/instrument/{cntnam}/data",
                    shape=cnt_shape,
                    dtype=cnt.dtype,
                )

                for i in range(cnt_shape[0]):
                    # print(np.unravel_index(i, A.shape), layout.shape, vsource[A.flat[i]].shape)

                    layout[(slice(None), *np.unravel_index(i, A.shape))] = vsource[
                        A.flat[i]
                    ]
                    layoutpymca[(*np.unravel_index(i, A.shape), slice(None))] = vsource[
                        A.flat[i]
                    ]

                if "data" in grp:
                    del grp["data"]

                if "pymca" in grp:
                    del grp["pymca"]

                grp.create_virtual_dataset("data", layout, fillvalue=-1)
                grp.create_virtual_dataset("pymca", layoutpymca, fillvalue=-1)
